//////////////////////////////////////////////////////////////
// La communication venant du master est geree par le hardware
// La communication venant des octaves suivant est generee par SoftwareSerial

// De maniere alternative on pourrait par le hardware , cela permettrait de laisser le hardware
// s'occuper de la partie intense en communications et le retour vers le master, plus simple, gere
// par le soft

// Tous les messages terminent par '\n'

// Reset general: recu du master: si le message commence par 'R', si c'est une erreur de comm tout
//                est resete ex ... si le master envoit 'RRRRR' tout est sur d'etre remis a 0, reponse
//                par un "R"+numOctave+"RR\n" vers le master pour signaler
// Assignation du numero d'octave: recu du master: si le message commence par 'S', le numero qui
//                suit est le numero d'octave courant (n) et 'S'+(n+1)+'\n' est transmit et
//                'S'+n+'\n' est renvoye au master
// Demande de l'etat des touches: recu du master : precede du numero d'octave, puis 'Q', reponse par
//                un message de la forme numOctave+'T'+etat des touches en hexadecimal dans un string
// Demande de d'allumage de touche: recu du master : precede du numero d'octave, puis 'O' suivi du
//                numero de la touche et du code de la couleur a afficher ('0' pas de couleur), reponse
//                par un message de la forme numOctave+'K'+numero de la touche+code de la couleur
//////////////////////////////////////////////////////////////

#define HARD_SPEED 38400
#define SOFT_SPEED 38400
#define MIDI_SPEED 31250


#include <SoftwareSerial.h>
#include <SPI.h>
#include <Wire.h>


#define debug Serial
bool bFirst = true;

SoftwareSerial toSlaveSerial( 4, 5 ); // RX, TX

byte touche[ 12 ] = { 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0 };

byte stateReturnPath;
char msgFromSlave[ 100 ];
int valueInKeyboardMessage;
int changesInProcess;

byte touchStatus[ 8 ][ 12 ];
byte allInstr[ 8 ][ 12 ];
byte allNotes[ 8 ][ 12 ];
byte allChanl[ 8 ][ 12 ];
byte allBanks[ 8 ][ 12 ];
byte allColor[ 8 ][ 12 ];

byte octaves = 0;

#define VS_XCS    6 // Control Chip Select Pin (for accessing SPI Control/Status registers)
#define VS_XDCS   7 // Data Chip Select / BSYNC Pin
#define VS_DREQ   2 // Data Request Pin: Player asks for more data

#define VS_RESET  8 //Reset is active low

//////////////////////////////////////////////////////////////
//Write to VS10xx register
//SCI: Data transfers are always 16bit. When a new SCI operation comes in
//DREQ goes low. We then have to wait for DREQ to go high again.
//XCS should be low for the full duration of operation.
void VSWriteRegister( unsigned char addressbyte, unsigned char highbyte, unsigned char lowbyte )
{
while ( !digitalRead( VS_DREQ ) ); //Wait for DREQ to go high indicating IC is available
digitalWrite( VS_XCS, LOW ); //Select control

//SCI consists of instruction byte, address byte, and 16-bit data word.
SPI.transfer( 0x02 ); //Write instruction
SPI.transfer( addressbyte );
SPI.transfer( highbyte );
SPI.transfer( lowbyte );
while ( !digitalRead( VS_DREQ ) ); //Wait for DREQ to go high indicating command is complete
digitalWrite( VS_XCS, HIGH ); //Deselect Control
}


//////////////////////////////////////////////////////////////
// Plugin to put VS10XX into realtime MIDI mode
// Originally from http://www.vlsi.fi/fileadmin/software/VS10XX/vs1053b-rtmidistart.zip
// Permission to reproduce here granted by VLSI solution.
//
const unsigned short sVS1053b_Realtime_MIDI_Plugin[ 28 ] = {
0x0007, 0x0001, 0x8050, 0x0006, 0x0014, 0x0030, 0x0715, 0xb080,
0x3400, 0x0007, 0x9255, 0x3d00, 0x0024, 0x0030, 0x0295, 0x6890,
0x3400, 0x0030, 0x0495, 0x3d00, 0x0024, 0x2908, 0x4d40, 0x0030,
0x0200, 0x000a, 0x0001, 0x0050,
};

//////////////////////////////////////////////////////////////
void VSLoadUserCode( void )
{
int i = 0;

while ( i < sizeof( sVS1053b_Realtime_MIDI_Plugin ) / sizeof( sVS1053b_Realtime_MIDI_Plugin[ 0 ] ) )
    {
    unsigned short addr, n, val;
    addr = sVS1053b_Realtime_MIDI_Plugin[ i++ ];
    n = sVS1053b_Realtime_MIDI_Plugin[ i++ ];
    while ( n-- )
        {
        val = sVS1053b_Realtime_MIDI_Plugin[ i++ ];
        VSWriteRegister( addr, val >> 8, val & 0xFF );
        }
    }
}

//////////////////////////////////////////////////////////////
void setupMidi( )
{
    // Volume - don't comment out this code!
talkMIDI( 0xB0, 0x07, 127 ); //0xB0 is channel message, set channel volume to max (127)

// ---------------------------------------------------------------------------------------------------------
// Melodic Instruments GM1
// ---------------------------------------------------------------------------------------------------------
// To Play "Electric Piano" (5):
talkMIDI( 0xB0, 0, 0x0 ); // Default bank GM1

// We change the instrument by changing the middle number in the brackets
// talkMIDI(0xC0, number, 0); "number" can be any number from the melodic table below
talkMIDI( 0xC0, 5, 0 ); // Set instrument number. 0xC0 is a 1 data byte command(55,0)
// ---------------------------------------------------------------------------------------------------------
// Percussion Instruments (Drums, GM1 + GM2)
// ---------------------------------------------------------------------------------------------------------
// uncomment the two lines of code below to use - you will also need to comment out the two "talkMIDI" lines
// of code in the Melodic Instruments section above
// talkMIDI(0xB0, 0, 0x78); // Bank select: drums
// talkMIDI(0xC0, 0, 0); // Set a dummy instrument number
// ---------------------------------------------------------------------------------------------------------

}

//////////////////////////////////////////////////////////////
void sendMIDI( byte data )
{
SPI.transfer( 0 );
SPI.transfer( data );
}

//////////////////////////////////////////////////////////////
//Plays a MIDI note. Doesn't check to see that cmd is greater than 127, or that data values are less than 127
void talkMIDI( byte cmd, byte data1, byte data2 )
{
    //
    // Wait for chip to be ready (Unlikely to be an issue with real time MIDI)
    //
while ( !digitalRead( VS_DREQ ) )
    ;
digitalWrite( VS_XDCS, LOW );

sendMIDI( cmd );
//Some commands only have one data byte. All cmds less than 0xBn have 2 data bytes 
//(sort of: http://253.ccarh.org/handout/midiprotocol/)
if ( (cmd & 0xF0) <= 0xB0 || (cmd & 0xF0) >= 0xE0 )
    {
    sendMIDI( data1 );
    sendMIDI( data2 );
    }
else
    {
    sendMIDI( data1 );
    }

digitalWrite( VS_XDCS, HIGH );
}

//////////////////////////////////////////////////////////////
//Send a MIDI note-on message.  Like pressing a piano key
//channel ranges from 0-15
void noteOn( byte octave, byte touche, byte attack_velocity )
{
debug.print( "instr " );
debug.println( allInstr[ octave ][ touche ] );
debug.print( "note " );
debug.println( allNotes[ octave ][ touche ] );
talkMIDI( (0xC0 | allChanl[ octave ][ touche ]), allInstr[ octave ][ touche ], 0 ); // Set instrument number. 0xC0 is a 1 data byte command(55,0)
talkMIDI( (0x90 | allChanl[ octave ][ touche ]), allNotes[ octave ][ touche ], attack_velocity );
}

//////////////////////////////////////////////////////////////
//Send a MIDI note-off message.  Like releasing a piano key
void noteOff( byte octave, byte touche, byte release_velocity )
{
talkMIDI( (0x80 | allChanl[ octave ][ touche ]), allNotes[ octave ][ touche ], release_velocity );
}

//////////////////////////////////////////////////////////////
void setup( )
{
delay( 800 );
Serial.begin( HARD_SPEED );
while ( !Serial );
toSlaveSerial.begin( SOFT_SPEED );
while ( !toSlaveSerial);

pinMode( VS_DREQ, INPUT );
pinMode( VS_XCS, OUTPUT );
pinMode( VS_XDCS, OUTPUT );
digitalWrite( VS_XCS, HIGH ); //Deselect Control
digitalWrite( VS_XDCS, HIGH ); //Deselect Data

pinMode( VS_RESET, OUTPUT );

debug.println( "MP3 Shield init - midi mode" );

//Initialize VS1053 chip 
digitalWrite( VS_RESET, LOW ); //Put VS1053 into hardware reset

//Setup SPI for VS1053
pinMode( 10, OUTPUT ); //Pin 10 must be set as an output for the SPI communication to work
SPI.begin( );
SPI.setBitOrder( MSBFIRST );
SPI.setDataMode( SPI_MODE0 );

//From page 12 of datasheet, max SCI reads are CLKI/7. Input clock is 12.288MHz. 
//Internal clock multiplier is 1.0x after power up. 
//Therefore, max SPI speed is 1.75MHz. We will use 1MHz to be safe.
SPI.setClockDivider( SPI_CLOCK_DIV16 ); //Set SPI bus speed to 1MHz (16MHz / 16 = 1MHz)
SPI.transfer( 0xFF ); //Throw a dummy byte at the bus

delayMicroseconds( 1 );
digitalWrite( VS_RESET, HIGH ); //Bring up VS1053

VSLoadUserCode( );

/////////////////////////////
debug.println( "setup midi" );
setupMidi( );

/////////////////////////////
debug.println( "setup touches" );
for ( int i = 0; i < 8; i++ )
    for ( int j = 0; j < 12; j++ )
        touchStatus[ i ][ j ] = 0;

/////////////////////////////
debug.println( "setup keypad" );
changesInProcess = 0;

/////////////////////////////
// give a chance to slave to finish their setup too
delay( 1000 );

debug.println( "reset" );
toSlaveSerial.write( "RRRRR\n" );
//waitFeedback( 1000 );

first( );
debug.println( "fin setup" );

}

//////////////////////////////////////////////////////////////
void reset( )
{
while ( toSlaveSerial.available( ) ) toSlaveSerial.read( );
while ( Serial.available( ) ) Serial.read( );
stateReturnPath = 0;

toSlaveSerial.write( "RRRRR\n" );
debug.println( "->RRRRR" );
}

//////////////////////////////////////////////////////////////
void resendAll( )
{
while ( toSlaveSerial.available( ) ) toSlaveSerial.read( );

toSlaveSerial.write( "TTTTT\n" );
debug.println( "->TTTTT" );
stateReturnPath = 0;
}

//////////////////////////////////////////////////////////////
// light a key - for midi playback
void ledOn( char octave, char touche )
{
char msg[ 10 ];
msg[ 0 ] = 'O';
msg[ 1 ] = 0x30 + octave;
msg[ 2 ] = 0x30 + touche;
msg[ 3 ] = 0x30 + allColor[ octave ][ touche ];
msg[ 4 ] = '\n';
msg[ 5 ] = 0;
toSlaveSerial.write( msg );
}

//////////////////////////////////////////////////////////////
// light a key - for midi playback
void ledOff( char octave, char touche )
{
char msg[ 10 ];
msg[ 0 ] = 'O';
msg[ 1 ] = 0x30 + octave;
msg[ 2 ] = 0x30 + touche;
msg[ 3 ] = 0x30; // color = 0 - leds off
msg[ 4 ] = '\n';
msg[ 5 ] = 0;
toSlaveSerial.write( msg );
}

//////////////////////////////////////////////////////////////
// global change of color for a key (at init)
void setLedColor( char octave, char touche, char color )
{
char msg[ 10 ];
msg[ 0 ] = 'P';
msg[ 1 ] = 0x30 + octave;
msg[ 2 ] = 0x30 + touche;
msg[ 3 ] = 0x30 + color;
msg[ 4 ] = '\n';
msg[ 5 ] = 0;
toSlaveSerial.write( msg );
}

//////////////////////////////////////////////////////////////

void messageFromSlave( char newIn )
{
if ( newIn == '$' ) newIn = '\n';

msgFromSlave[ stateReturnPath++ ] = newIn;
msgFromSlave[ stateReturnPath ] = 0;
if ( newIn == '\n' )
    { // fin de message
    debug.print( msgFromSlave );
    switch ( msgFromSlave[ 0 ] )
        {
        //  got from slaves
        case 'd': // nothing to do, already displayed
        case 'D': // nothing to do, already displayed
        //Serial.write( msgReturnPath, stateReturnPath );
            break;
        case 'n': // Note status changed on slave
        {
        int octave = msgFromSlave[ 1 ] - 0x30;
        int touche = msgFromSlave[ 2 ] - 0x30;
        int status = msgFromSlave[ 3 ] - 0x30;
        if ( octave > 7 || touche > 12 || status > 2 )
            { // Wrong value
            reset( );
            debug.println( "reset fromserver, invalid data" );
            }
        else
            { // Manage touch status
            if ( status == 1 )
                { // pressed
                if ( changesInProcess != 0 )
                    {
                    byte octaveTmp = octave;
                    byte toucheTmp = touche;
                    debug.println( valueInKeyboardMessage );
                    debug.println( octave );
                    debug.println( touche );
                    switch ( changesInProcess )
                        {
                        case 1://change instrument
                            debug.println( "instr " );
                            while ( octaveTmp < 8 )
                                {
                                while ( toucheTmp < 12 )
                                    {
                                    allInstr[ octaveTmp ][ toucheTmp ] = valueInKeyboardMessage;
                                    toucheTmp++;
                                    }
                                toucheTmp = 0;
                                octaveTmp++;
                                }
                            changesInProcess = 0;
                            break;
                        case 2://change octave
                            debug.println( "transpo" );
                            while ( octaveTmp < 8 )
                                {
                                while ( toucheTmp < 12 )
                                    {
                                    allNotes[ octaveTmp ][ toucheTmp ] = (valueInKeyboardMessage + 1) * 12 + toucheTmp;
                                    allChanl[ octaveTmp ][ toucheTmp ]++;
                                    toucheTmp++;
                                    }
                                toucheTmp = 0;
                                octaveTmp++;
                                valueInKeyboardMessage++;
                                }
                            changesInProcess = 0;
                            break;
                        case 3://change color
                            changesInProcess = 0;
                            break;
                        case 4://change bank
                            changesInProcess = 0;
                            break;
                        }
                    }
                noteOn( octave, touche, 0x60 );
                touchStatus[ octave ][ touche ] = 1;
                }
            else if ( status == 2 )
                { // released
                noteOff( octave, touche, 0x60 );
                touchStatus[ octave ][ touche ] = 0;
                }
            }
        }
        break;
        case 'o':
        case 'p': //just feedback
            break;
        case 'r': // get maximum of octaves (from s)
        // use r too by security but should be needed
        //break;
        case 's':
            //octaves = max( octaves, msgReturnPath[ 1 ] - 0x30 );
            break;
        case 't':
        {
        int j = 1;
        int oct = msgFromSlave[ j++ ] - 0x30;
        for ( int i = 0; i < 12; i++ )
            {
            touchStatus[ oct ][ i ] = msgFromSlave[ j++ ] - 0x30;
            }
        }
        break;
        // configured on master
        case 'K': // use Keyboard to change instrument/octave/other
        {
        int idx = 1;
        switch ( msgFromSlave[ idx++ ] )
            {
            case 'A'://changement d'instrument
                changesInProcess = 1;
                valueInKeyboardMessage = atoi( &(msgFromSlave[ idx ]) );
                break;
            case 'B'://changement d'octave
                changesInProcess = 2;
                valueInKeyboardMessage = atoi( &(msgFromSlave[ idx ]) );
                break;
            case 'C'://changement de couleur de la touche
                changesInProcess = 3;
                valueInKeyboardMessage = atoi( &(msgFromSlave[ idx ]) );
                break;
            case 'D'://changement de banque midi
                changesInProcess = 4;
                valueInKeyboardMessage = atoi( &(msgFromSlave[ idx ]) );
                break;
            }

        }
        break;
        default:
            debug.println( "---->> unknown code\n" );
            
            break;
        }
    stateReturnPath = 0;
    }
}

int waitFeedback( int valDelay )
{
int i = 0;
bool bStop = false;
// wait for return
while ( !bStop && !toSlaveSerial.available( ) )
    {
    delay( 1 );
    i++;
    if ( i > valDelay ) bStop = true;
    }
if ( !bStop )
    {
    while ( toSlaveSerial.available( ) )
        {
        messageFromSlave( toSlaveSerial.read( ) );
        }
    return 1;
    }
debug.println( "wait failed" );
return 0;
}


//////////////////////////////////////////////////////////////
void first( )
{
bool bEnd = false;
char msg[ 10 ];
octaves = 0;
msg[ 0 ] = 'S';
msg[ 1 ] = 0x30+octaves;
msg[ 2 ] = '\n';
msg[ 3 ] = 0;
debug.print( "Init Octaves\n" );
while ( bEnd == false )
    {
    debug.print( msg );
    toSlaveSerial.write( msg );
    if ( !waitFeedback( 1000 ) )
        {
        bEnd = true;
        }
    else
        {
        debug.println( msgFromSlave );
        if ( msgFromSlave[0] == 's' )
            {
            octaves++;
            msg[ 1 ] = octaves + 0x30;
            }
        }
    }

debug.print( "octaves : " );
debug.println( octaves );


// here, I should know how many octaves are ready.
int note = 0;
if ( octaves > 7 ) octaves = 2;

switch ( octaves )
    {
    case 1:
        note = 60;
        break;
    case 2:
        note = 60;
        break;
    case 3:
        note = 48;
        break;
    case 4:
        note = 48;
        break;
    case 5:
        note = 36;
        break;
    case 6:
        note = 36;
        break;
    case 7:
        note = 24;
        break;
    case 8:
        note = 24;
        break;
    default:
        note = 48;
        break;
    }
int colordn = 1;
int colorup = 1;

debug.print( "Found " );
debug.print( octaves );
debug.println( " octaves" );

// init Colors
debug.println( "Init colors" );
for ( int i = 0; i < octaves; i++ )
    {
    int colordn = 3;
    int colorup = 1;
    for ( int j = 0; j < 12; j++ )
        {
        if ( touche[ j ] == 1 )
            {
            allColor[ i ][ j ] = colorup;
            colorup++;
            if ( colorup > 5 ) colorup = 1;
            }
        else
            {
            allColor[ i ][ j ] = colordn;
            colordn--;
            if ( colordn < 1 ) colordn = 7;
            }
            //allColor[i][j] = 7;

        setLedColor( i, j, allColor[ i ][ j ] );
        //waitFeedback(0);

        // set array
        allInstr[ i ][ j ] = 1;
        allChanl[ i ][ j ] = 0;
        allNotes[ i ][ j ] = note;
        //Note on channel 1 (0x90), some note value (note), middle velocity (0x45):

        noteOn( i, j, 127 );
        ledOn( i, j );
        delay( 200 );

        //Turn off the note with a given off/release velocity
        noteOff( i, j, 127 );
        ledOff( i, j );
        note++;
        }
    }

    /*    for ( int i = 0; i < octaves; i++ ) {
            for ( int j = 0; j < 12; j++ ) {
            //Note on channel 1 (0x90), some note value (note), middle velocity (0x45):
            ledOn( i, j );
            delay( 200 );
            ledOff( i, j );
            }
            }*/

    /*    // init Notes
        debug.println( "Init keys" );
        for ( int i = 0; i < octaves; i++ ) {
        for ( int j = 0; j < 12; j++ ) {
        //Note on channel 1 (0x90), some note value (note), middle velocity (0x45):
        allNotes[i][j] = note;
        noteOn( i, j, 127 );
        ledOn( i, j );
        delay( 200 );

        //Turn off the note with a given off/release velocity
        noteOff( i, j, 127 );
        ledOff( i, j );
        note++;
        }
        }
        */
    //for ( int i = 0; i < octaves; i++ ) {
    //	for ( int j = 0; j < 12; j++ ) {
    //		ledOn( i, j );
    //	}
    //}

    //delay( 200 );
    /*
for (int z=0; z < 2; z++)
*/
for ( int i = 0; i < octaves; i++ )
    {
    for ( int j = 0; j < 12; j++ )
        {
        ledOff( i, j );
        }
    }

Serial.println( "Note initialized" );

bFirst = false;
}

//////////////////////////////////////////////////////////////
void loop( )
{
//if ( bFirst == true ) first( );

if ( toSlaveSerial.available( ) )
    {
    messageFromSlave( toSlaveSerial.read( ) );
    }

if ( Serial.available( ) )
    {
    messageFromSlave( Serial.read( ) );
    }
}
